#pragma once
#include "grbc/ext.h"
#include <sol/table.hpp>
#include <sol/types.hpp>
#include <string>
#include <vector>

#define GRBC_PROPERTY_CXX_VERSION "CXX_VERSION"

// https://stackoverflow.com/questions/1505582/determining-32-vs-64-bit-in-c

#if _WIN32 || _WIN64
#if _WIN64
#define ENVIRONMENT64
#else
#define ENVIRONMENT32
#endif
#endif

// Check GCC
#if __GNUC__
#if __x86_64__ || __ppc64__
#define ENVIRONMENT64
#else
#define ENVIRONMENT32
#endif
#endif

#define GRBC_VERSION "1.0"

struct Property {
  Property() = default;

  Property(const sol::table &table) {
    name = table.get<std::string>("name");
    value = table.get<std::string>("value");
  }

  /// Name of this property
  std::string name;

  /// Value of this property
  std::string value;
};

enum LanguageType { LanguageType_CPP, LanguageType_C };
enum LibraryType { LibraryType_Shared, LibraryType_Static };

struct TargetInfo {
  /// Name of the target
  std::string name;
};

struct Package {
  Package(const sol::table &table) {
    name = table.get<std::string>("name");
    compiler_flags = table.get<std::string>("compiler_flags");
    linker_flags = table.get<std::string>("linker_flags");
  }

  Package() = default;

  std::string name;

  std::string file_name;

  std::string compiler_flags;

  std::string linker_flags;
};

struct PackageConfig {
  PackageConfig(const sol::table &table) {
    name = table.get<std::string>("name");
    libraries = table.get<std::vector<Package>>("libraries");
    include_dirs = table.get<std::vector<std::string>>("include_dirs");
    compile_flags = table.get<std::vector<std::string>>("compile_flags");
    linker_flags = table.get<std::vector<std::string>>("linker_flags");
  }

  PackageConfig() = default;

  std::string name;

  std::vector<Package> libraries;

  std::vector<std::string> include_dirs;

  std::vector<std::string> compile_flags;

  std::vector<std::string> linker_flags;
};

enum PlatformType { PlatformType_Win32, PlatformType_Linux };

struct GlobalConfig {
  /// Version of the grbc executable
  std::string engine_version = GRBC_VERSION;

  /// Architecture string
  std::string architecture;

  /// Build directory
  std::string build_dir = "build";

  /// Path to the platform.hcfg, in_memory if not on-disk
  std::string platform_config;
};

struct ExecutableConfig {
  ExecutableConfig(const sol::table &table) {
    name = table.get<std::string>("name");
    files = table.get<std::vector<std::string>>("files");

    requirements = table.get<std::vector<Package>>("requirements");

    compile_flags = table.get<std::vector<std::string>>("compile_flags");

    linker_flags = table.get<std::vector<std::string>>("linker_flags");

    include_dirs = table.get<std::vector<std::string>>("include_dirs");

    language_type = table.get<LanguageType>("language_type");

    if (table["properties"].valid()) {
      properties = table.get<std::vector<Property>>("properties");
    }
  }

  /// Name of the executable
  std::string name;

  /// Type of language
  LanguageType language_type;

  /// List of files to compile
  std::vector<std::string> files{};

  /// Requirments of the executable
  std::vector<Package> requirements{};

  /// Compiler flags
  std::vector<std::string> compile_flags{};

  /// Linker flags
  std::vector<std::string> linker_flags{};

  /// Include directories
  std::vector<std::string> include_dirs{};

  /// Optional list of properties
  std::vector<Property> properties;
};

struct TaskConfig {
  TaskConfig() = default;
  TaskConfig(const sol::table &table) {
    name = table.get<std::string>("name");
    task_id = table.get<std::string>("task_id");
    shell_script = table.get<std::string>("shell_script");
  }

  /// Name of the task
  std::string name;

  /// ID of the task
  std::string task_id;

  /// Shell script content for this task
  std::string shell_script;
};

struct LibraryConfig {
  LibraryConfig(const sol::table &table) {
    name = table.get<std::string>("name");
    files = table.get<std::vector<std::string>>("files");

    requirements = table.get<std::vector<Package>>("requirements");

    compile_flags = table.get<std::vector<std::string>>("compile_flags");

    linker_flags = table.get<std::vector<std::string>>("linker_flags");

    include_dirs = table.get<std::vector<std::string>>("include_dirs");

    language_type = table.get<LanguageType>("language_type");
    lib_type = table.get<LibraryType>("lib_type");

    if (table["package_config"].valid())
      package_config = table.get<PackageConfig>("package_config");

    if (table["properties"].valid())
      properties = table.get<std::vector<Property>>("properties");
  }

  /// Name of the executable
  std::string name;

  /// Type of language
  LanguageType language_type;

  /// Type of library
  LibraryType lib_type;

  /// Package config
  PackageConfig package_config{};

  /// List of files to compile
  std::vector<std::string> files{};

  /// Requirments of the executable
  std::vector<Package> requirements{};

  /// Compiler flags
  std::vector<std::string> compile_flags{};

  /// Linker flags
  std::vector<std::string> linker_flags{};

  /// Include directories
  std::vector<std::string> include_dirs{};

  /// Optional list of properties
  std::vector<Property> properties;
};

struct Platform {
  /// Name of the platform
  std::string name;

  /// C++ compiler
  std::string cxx_compiler;

  /// C compiler
  std::string cc_compiler;

  /// Is this platform 64-bit?
  bool is_64bit = true;

  /// Type of the platform
  PlatformType platform_type = PlatformType_Linux;
};

/// Functions

void grbc_want_version(const std::string &version);

void grbc_exception(const std::string &exception_string);

GlobalConfig &grbc_get_config();

std::string grbc_file(const std::string &file_path);

TargetInfo grbc_executable(const ExecutableConfig &executable_config);

TargetInfo grbc_library(const LibraryConfig &library_config);

void grbc_load_platform(const std::string &file_path);

void grbc_set_platform(const Platform &platform);

std::string grbc_find_compiler(const std::string &compiler_name);

void grbc_build(const std::string &generator_id);

std::string grbc_object_file(const std::string &file_path);

std::string grbc_replace_string(const std::string &string, char substring,
                                char replacement);

std::string
grbc_include_dirs_to_cflags(const std::vector<std::string> &include_dirs);

void grbc_log(const std::string &message);

bool grbc_is_win32();

bool grbc_is_linux();

PlatformType grbc_get_platform();

bool grbc_is_64bit();

bool grbc_is_32bit();

std::string grbc_compiler_define(const std::string &define,
                                 const std::string &value);

std::string grbc_get_lib_extension(LibraryType lib_type);

Package grbc_bake_package_config(const PackageConfig &config);

Package grbc_pkg(const std::string &package_name);

void grbc_ext(const std::string &extension_id);

void grbc_register_ext(const Extension &ext);

bool grbc_has_ext(const std::string &extension_id);

std::string grbc_platform_file(PlatformType platform_type,
                               const std::string &file_name);

bool grbc_is_ext_loaded(const std::string &extension_name);

void grbc_task(const TaskConfig &config);

bool grbc_has_option(const std::string &option_name);

std::string grbc_get_option(const std::string &option_name);

Property grbc_property(const std::string &key, const std::string &value);

Property grbc_cxx_version(const std::string &version_string);

void grbc_global_properties(const sol::table &properties);